------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Config;
use Config;

with Tables;

with Scheduling_Sequences;
use Scheduling_Sequences;

with Time_Units;
use Time_Units;


with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Dependencies;
use Dependencies;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Unbounded_Strings;
use Unbounded_Strings;



with Unchecked_Deallocation;




package Buffers is


   type Role is 
         (No_Role,  
          Producer, 
          Consumer); 
   type Buffer_Role is 
      record 
         The_Role : Role;  
         Size     : Natural;  
      end record; 

   procedure Put (
         B : in     Buffer_Role ); 
   procedure Initialize (
         B :    out Buffer_Role ); 

   package Buffer_Roles is new Tables(Buffer_Role, Unbounded_String,
      Config.Max_Tasks, 0,
      Put, Initialize, Unbounded_Strings.Put);
   use Buffer_Roles;

   subtype Buffer_Roles_Table is Table;
   subtype Buffer_Roles_Range is Table_Range;

   --------------------------------------------------------

   type Buffer is new Dependency with 
      record 
         Name     : Unbounded_String   := To_Unbounded_String ("");  
         Location : Unbounded_String   := To_Unbounded_String ("");  
         Size     : Natural            := 0;  
         Roles    : Buffer_Roles_Table;  
      end record; 

   type Buffer_Ptr is access all Buffer'Class; 


   procedure Initialize (
         A_Buff : in out Buffer ); 
   procedure Finalize (
         A_Buff : in out Buffer ); 
   procedure Adjust (
         A_Buff : in out Buffer ); 


   procedure Dispatched_Dump (
         My_Buff : in     Buffer ); 

   function Dispatched_Copy (
         My_Buffer : Buffer ) 
     return Dependency_Ptr; 



   procedure Free is 
   new Unchecked_Deallocation
      (Buffer'Class, Buffer_Ptr);


   procedure Dump (
         My_Buff : in     Buffer_Ptr ); 

   procedure Put_Name (
         My_Buff : in     Buffer_Ptr ); 


   function Copy (
         My_Buffer : in     Buffer_Ptr ) 
     return Buffer_Ptr; 







   --------------------------------------------------------


   type Buffer_Size is 
      record 
         Time : Natural;  
         Size : Natural;  
      end record; 
   type Buffer_Size_Table is array (Time_Unit_Range) of Buffer_Size; 
   procedure Initialize (
         B :    out Buffer_Size ); 
   procedure Initialize (
         B :    out Buffer_Size_Table ); 



   function Get_Max_Size (
         Buff : Buffer_Size_Table ) 
     return Natural; 

   procedure Get_Buffer_Size (
         Sched    : in     Scheduling_Sequence; 
         My_Roles : in     Buffer_Roles_Table;  
         My_Tasks : in     Tasks_Set;           
         Buff     : in out Buffer_Size_Table    ); 

   --------------------------------------------------------

   function Get_Bound (
         My_Roles       : in     Buffer_Roles_Table; 
         My_Tasks       : in     Tasks_Set;          
         Processor_Name : in     Unbounded_String    ) 
     return Natural; 


   function Get_Bound_Harmonic (
         My_Roles       : in     Buffer_Roles_Table; 
         My_Tasks       : in     Tasks_Set;          
         Processor_Name : in     Unbounded_String    ) 
     return Natural; 

   function Is_Cons_Prod_Harmonic (
         My_Tasks : in     Tasks_Set;         
         My_Roles : in     Buffer_Roles_Table ) 
     return Boolean; 

   Flow_Constraint_Not_Respected : exception;  

   procedure Buffer_Flow_Control (
         My_Roles : in     Buffer_Roles_Table; 
         My_Tasks : in     Tasks_Set           ); 




end Buffers;





