------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Text_Io;
use Text_Io;

use Processor_Set.Generic_Processor_Set;

with Scheduler.Edf;
use Scheduler.Edf;
with Scheduler.Llf;
use Scheduler.Llf;
with Scheduler.Fixed_Priority.Dm;
use Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm;
use Scheduler.Fixed_Priority.Rm;
with Scheduler.Fixed_Priority.Hpf;
use Scheduler.Fixed_Priority.Hpf;





package body Processor_Set is




   procedure Fill_Descriptor (
         My_Processor : in     Processors_Set;       
         Desc         :    out Processor_Descriptor; 
         A_Processor  : in     Processor_Ptr         ) is 

   begin

      Desc.Quantum:=Get_Quantum(A_Processor.Scheduler.all);
      Desc.Processor_Name := To_Bounded_String(To_String(A_Processor.Name));
      Desc.Is_Preemptive:=Get_Preemptive(A_Processor.Scheduler.all);
      Desc.Scheduler_Name:=Get_Name(A_Processor.Scheduler.all);

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Processor : in     Processors_Set;       
         Desc         : in     Processor_Descriptor; 
         A_Processor  :    out Processor_Ptr         ) is 

      Scheduler : Scheduler_Ptr;  

   begin
      A_Processor := new Processor;

      if (Desc.Scheduler_Name = Earliest_Deadline_First)
            then
         Scheduler:= new Edf_Scheduler;
      else
         if (Desc.Scheduler_Name = Least_Laxity_First)
               then
            Scheduler:= new Llf_Scheduler;
         else
            if (Desc.Scheduler_Name = Rate_Monotonic)
                  then
               Scheduler:= new Rm_Scheduler;
            else
               if (Desc.Scheduler_Name = Deadline_Monotonic)
                     then
                  Scheduler:= new Dm_Scheduler;
               else
                  if (Desc.Scheduler_Name = Highiest_Priority_First)
                        then
                     Scheduler:= new Hpf_Scheduler;
                  end if;
               end if;
            end if;
         end if;
      end if;


      Set_Preemptive(Scheduler.all, Desc.Is_Preemptive);
      Set_Quantum(Scheduler.all, Desc.Quantum);

      A_Processor.Scheduler:=Scheduler;
      A_Processor.Name:=To_Unbounded_String(To_String(Desc.Processor_Name));


   end Extract_Descriptor;


   procedure Add_Processor (
         My_Processors : in out Processors_Set;   
         Name          : in     Unbounded_String; 
         A_Scheduler   : in     Scheduler_Ptr     ) is 

      My_Iterator : Iterator;  
      A_Processor : Processor_Ptr;  

   begin

      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;
      if (A_Scheduler = null)
            then
         raise Invalid_Parameter;
      end if;


      if(Get_Number_Of_Elements(My_Processors) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Processors, A_Processor, My_Iterator);
            if (Name = A_Processor.Name)
                  then
               raise Processor_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Processors, My_Iterator) = True;

            Next_Element(My_Processors, My_Iterator);


         end loop;

      end if;


      A_Processor:= new Processor;
      A_Processor.Name := To_Unbounded_String(To_String(Name));
      A_Processor.Scheduler := A_Scheduler;

      Add(My_Processors, A_Processor);


   end Add_Processor;




   function Get (
         My_Processors  : in     Processors_Set;      
         Processor_Name : in     Unbounded_String;    
         Param_Name     : in     Processor_Parameters ) 
     return Scheduler_Ptr is 

      A_Processor : Processor_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Scheduler)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processors,A_Processor,My_Iterator);

         if (A_Processor.Name = Processor_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Processors, My_Iterator) = True;

         Next_Element(My_Processors, My_Iterator);

      end loop;

      return A_Processor.Scheduler;

   end Get;



   procedure Set (
         My_Processors  : in out Processors_Set;       
         Processor_Name : in     Unbounded_String;     
         Param_Name     : in     Processor_Parameters; 
         Param_Value    : in     Scheduler_Ptr         ) is 

      A_Processor : Processor_Ptr;  
      My_Iterator : Iterator;  
   begin

      if (Param_Name /= Scheduler)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Processors,A_Processor,My_Iterator);

         if (A_Processor.Name = Processor_Name)
               then
            A_Processor.Scheduler:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Processors, My_Iterator) = True;

         Next_Element(My_Processors, My_Iterator);

      end loop;
   end Set;



end Processor_Set;
