------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Text_Io;
use Text_Io;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;


package body Scheduler.Fixed_Priority.Hpf is


   procedure Initialize (
         A_Scheduler : in out Hpf_Scheduler ) is 
   begin
      Reset(A_Scheduler);
      A_Scheduler.Name:= Highiest_Priority_First;
   end Initialize;

   procedure Finalize (
         A_Scheduler : in out Hpf_Scheduler ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         A_Scheduler : in out Hpf_Scheduler ) is 
   begin
      null;
   end Adjust;



   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Hpf_Scheduler;   
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

   begin

      null;

   end Check_Before_Scheduling;




   procedure Dispatched_Response_Time (
         My_Scheduler   : in     Hpf_Scheduler;    
         My_Tasks       : in out Tasks_Set;        
         Processor_Name : in     Unbounded_String; 
         Msg            : in out Unbounded_String; 
         Response_Time  :    out Response_Table    ) is 


   begin

      -- check if tasks are periodics 
      Periodic_Control(My_Tasks, Processor_Name);

      -- Check processor utilization first
      if (Processor_Utilization_Over_Period(My_Scheduler, My_Tasks,
               Processor_Name) > 1.0)
            then
         raise Processor_Utilization_Exceeded;
      end if;

      -- check start time
      Start_Time_Control(My_Tasks, Processor_Name);

      -- check offset
      Offset_Control(My_Tasks, Processor_Name);


      -- Compute response time
      --
      Scheduler.Fixed_Priority.Compute_Response_Time (
         My_Scheduler, My_Tasks, Processor_Name,
         Msg, Response_Time);


   end Dispatched_Response_Time;





   procedure Put_Tcb (
         E : Hpf_Tcb_Ptr ) is 
   begin
      null;
   end Put_Tcb;


   function Build_Tcb (
         My_Scheduler : in     Hpf_Scheduler; 
         A_Task       :        Task_Ptr       ) 
     return Tcb_Ptr is 

      A_Tcb : Hpf_Tcb_Ptr;  
   begin
      A_Tcb:=new Hpf_Tcb;

      A_Tcb.Tsk:=A_Task;
      A_Tcb.Activation:=1;
    A_Tcb.end_time:=0;
      A_Tcb.Cpu_Usage:=0;
      A_Tcb.Task_Quantum:=My_Scheduler.Quantum;
      A_Tcb.Wake_Up_Time:=A_Task.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Task.Capacity;

      return Tcb_Ptr(A_Tcb);
   end Build_Tcb;





   procedure Specific_Scheduler_Initialization (
         My_Scheduler : in out Hpf_Scheduler; 
   si : in out scheduling_information;
        Processor_Name    : in     Unbounded_String;
         My_Tasks     : in out Tasks_Set;     
         My_Resources : in out Resources_Set  ) is 


      Index : Tasks_Range := 0;  

   begin
      for I in Priority_Range loop
         Reset(My_Scheduler.Priority_Fifos(I));
      end loop;

      -- Insert task info fifos according to
      -- their priority level
      --
      loop
	if si.Tcbs(Index).Tsk.location = processor_name then 
         Insert(My_Scheduler.Priority_Fifos(si.Tcbs(Index).Tsk.
               Priority), Hpf_Tcb_Ptr(si.Tcbs(Index)));
	end if;

         Index:=Index+1;
         exit when si.Tcbs(Index)=null;
      end loop;

   end Specific_Scheduler_Initialization;




   procedure Do_Election (
         My_Scheduler      : in out Hpf_Scheduler;
  si : in out scheduling_information;
Processor_Name : in     Unbounded_String;
         At_Time           :        Natural;
My_Dependencies : in     Tasks_Dependencies;
         With_Offsets      :        Boolean;
         With_Precedencies :        Boolean;
         With_Resources    :        Boolean;
         Elected           : in out Tasks_Range;
         No_Task           : in out Boolean        ) is 


      Ready,  
      Found             : Boolean;  
      Selected_Priority : Priority_Range;  
      Index             : Tasks_Range    := 0;  
      Current           : Hpf_Tcb_Ptr;  
      Elected_Task      : Hpf_Tcb_Ptr    := null;  
      Max_Cpu_Usage     : Natural        := Natural'Last;  


   begin


      -- For each task, call "check_ressources" to
      -- take care of priority modification (PCP and PIP)
      --
      for I in 0..si.Number_Of_Tasks-1 loop
	if si.Tcbs(I).tsk.location = processor_name then
         Check_Resource(My_Scheduler, si, si.Tcbs(I),
            Fixed_Priority_Tcb_Ptr(si.Tcbs(I)).Is_Ready);
	end if;
      end loop;


      -- Scan fifos to find the highiest priority level
      -- with a ready task
      -- Warning : WE DO NOT CHANGE TASK POSITION IN THE 
      --	FIFOS 
      --
      Selected_Priority:=Priority_Range'Last;
      Found:=False;
      loop

         if not Is_Empty(My_Scheduler.Priority_Fifos(Selected_Priority))

               then

            for I in 1..Get_Size(My_Scheduler.Priority_Fifos(
                     Selected_Priority)) loop

               Current:=Consult(My_Scheduler.Priority_Fifos(
                     Selected_Priority));

               -- is the task ready ?
               --
               Ready:=False;
               if (Current.Wake_Up_Time <= At_Time) and
                     (Current.Rest_Of_Capacity /= 0)
                     then
                  if (With_Offsets=False) or
                        Check_Offset(Tcb_Ptr(Current), At_Time)
                        then
            if (With_precedencies=False) or
                  Check_precedencies(si, my_dependencies, at_time, Tcb_Ptr(Current))
                  then
                     if (With_Resources=False) or (Fixed_Priority_Tcb_Ptr(
                              Current).Is_Ready=True)
                           then

                        -- the priority level is selected
                        --
                        Ready:=True;
                        Found:=True;
                        exit;
                     end if;
                  end if;
                  end if;
               end if;

               -- The current task is not ready ...
               -- put it the queue of the fifo
               -- and look for the Next element 
               -- of the fifo
               --
               if (Ready=False)
                     then
                  Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Current);
                  Insert(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Current);
               end if;


            end loop;

         end if;



         -- A fifo is already selected ... get out !!
         -- and never come back .........
         --
         if (Found=True)
               then
            No_Task:=False;
            exit;
         end if;


         -- No task found = no task to scheduler;
         --
         if (Selected_Priority = Priority_Range'First)
               then
            No_Task:=True;
            return;
         end if;


         Selected_Priority:=
            Selected_Priority-1;

      end loop;






      -- Priority level is selected : find the task
      -- in the head of the fifo : this task is the elected
      -- task
      --

      -- Task sharing policies
      --
      if (Selected_Priority = 0)
            then
         -- Loop and looking for the less consuming
         -- task
         --
         Max_Cpu_Usage:=Natural'Last;
         for I in 1..Get_Size(My_Scheduler.Priority_Fifos(
                  Selected_Priority)) loop

            Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
               Current);
            Insert(My_Scheduler.Priority_Fifos(Selected_Priority), Current);
            if (Current.Wake_Up_Time <= At_Time) and
                  (Current.Rest_Of_Capacity /= 0)
                  then
               if (With_Offsets=False) or
                     Check_Offset(Tcb_Ptr(Current), At_Time)
                     then
                  if (With_Resources=False) or (Fixed_Priority_Tcb_Ptr(
                           Current).Is_Ready=True)
                        then

                     if (Current.Cpu_Usage <= Max_Cpu_Usage)
                           then
                        Max_Cpu_Usage:=Current.Cpu_Usage;
                        Elected_Task:=Current;
                     end if;
                  end if;
               end if;
            end if;
         end loop;



      else

         -- SCHED_FIFO or SCHED_RR task	
         --
         Elected_Task:=Consult(My_Scheduler.Priority_Fifos(
               Selected_Priority));

         -- is the task quantum exausted ?
         -- if so and if SCHED_RR is set
         -- the task have to be moved to 
         -- the queue of the fifo
         --

         if (Elected_Task.Tsk.Policy = Sched_Rr) then
            if (My_Scheduler.Quantum>0) then


               Elected_Task.Task_Quantum:=Elected_Task.Task_Quantum-1;

               if (Elected_Task.Task_Quantum=0)
                     then
                  Elected_Task.Task_Quantum:=My_Scheduler.Quantum;
                  Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Elected_Task );
                  Insert(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Elected_Task);
               end if;
            end if;
         end if;
      end if;



      if (Elected_Task /= null)
            then

         Index:=0;
         loop
            if (si.Tcbs(Index).Tsk.Name
                  = Elected_Task.Tsk.Name)
                  then
               Elected:=Index;
               exit;
            end if;

            Index:=Index+1;
            exit when si.Tcbs(Index)=null;
         end loop;

      end if;


      -- Update task information
      --
      Elected_Task.Cpu_Usage:=Elected_Task.Cpu_Usage+1;




   end Do_Election;







end Scheduler.Fixed_Priority.Hpf ;


