------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Config;
use Config;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Scheduler;
use Scheduler;


with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Unbounded_Strings;
use Unbounded_Strings;




package body Scheduler.Llf is


   procedure Initialize (
         A_Scheduler : in out Llf_Scheduler ) is 
   begin
      Reset(A_Scheduler);
      A_Scheduler.Name:= Least_Laxity_First;
   end Initialize;

   procedure Finalize (
         A_Scheduler : in out Llf_Scheduler ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         A_Scheduler : in out Llf_Scheduler ) is 
   begin
      null;
   end Adjust;





   function Build_Tcb (
         My_Scheduler : in     Llf_Scheduler; 
         A_Task       :        Task_Ptr       ) 
     return Tcb_Ptr is 

      A_Tcb : Tcb_Ptr;  
   begin
      A_Tcb:=new Llf_Tcb;

      A_Tcb.Tsk:=A_Task;
      A_Tcb.Activation:=1;
      A_Tcb.End_Time:=0;
      A_Tcb.Wake_Up_Time:=A_Task.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Task.Capacity;

      if A_Task.Task_Type = Periodic_Type
            then
         if(A_Task.Deadline /= 0)
               then
            Llf_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=
               A_Task.Deadline+A_Tcb.Wake_Up_Time;
         else
            Llf_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=Periodic_Task_Ptr(
               A_Task).Period+A_Tcb.Wake_Up_Time;
         end if;
      else
         Llf_Tcb_Ptr(A_Tcb).Dynamic_Deadline:=A_Task.Deadline;
      end if;

      return A_Tcb;
   end Build_Tcb;


   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Llf_Scheduler;   
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

   begin
      null;
   end Check_Before_Scheduling;


   procedure Do_Election (
         My_Scheduler      : in out Llf_Scheduler;          
         Si                : in out Scheduling_Information; 
         Processor_Name    : in     Unbounded_String;       
         At_Time           :        Natural;                
         My_Dependencies   : in     Tasks_Dependencies;     
         With_Offsets      :        Boolean;                
         With_Precedencies :        Boolean;                
         With_Resources    :        Boolean;                
         Elected           : in out Tasks_Range;            
         No_Task           : in out Boolean                 ) is 

      Smallest_Laxity : Natural     := Natural'Last;  
      Laxity          : Natural;  
      I               : Tasks_Range := 0;  
   begin

      loop
         if (Si.Tcbs(I).Tsk.Location = Processor_Name) then

            Laxity:=Llf_Tcb_Ptr(Si.Tcbs(I)).Dynamic_Deadline -
               Si.Tcbs(I).
               Rest_Of_Capacity;

            if (Si.Tcbs(I).Wake_Up_Time <= At_Time) and
                  (Laxity < Smallest_Laxity) and
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  then
               if (With_Offsets=False) or
                     Check_Offset(Si.Tcbs(I), At_Time)
                     then
                  if (With_Precedencies=False) or
                        Check_Precedencies(Si, My_Dependencies, At_Time,
                        Si.Tcbs(I))
                        then

                     Smallest_Laxity:=Laxity;
                     Elected:=I;
                  end if;
               end if;
            end if;
         end if;

         I:=I+1;
         exit when Si.Tcbs(I)=null;
      end loop;



      if Smallest_Laxity= Natural'Last
            then
         No_Task:=True;
      else
         No_Task:=False;

         -- last time unit : compute dynamic deadline for periodic task
         --
         if (Si.Tcbs(Elected).Rest_Of_Capacity = 1)
               and (Si.Tcbs(Elected).Tsk.Task_Type = Periodic_Type)
               then
            Llf_Tcb_Ptr(Si.Tcbs(Elected)).Dynamic_Deadline:=
               Llf_Tcb_Ptr(Si.Tcbs(Elected)).Dynamic_Deadline
               +Periodic_Task_Ptr(Si.Tcbs(Elected).Tsk).Period;
         end if;

      end if;

   end Do_Election;



   procedure Dispatched_Bound_On_Processor_Utilization (
         My_Scheduler   : in     Llf_Scheduler;    
         My_Tasks       : in     Tasks_Set;        
         Processor_Name : in     Unbounded_String; 
         Result         : in out Double;           
         Msg            : in out Unbounded_String  ) is 
   begin
      Result:=1.0;
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 8, ") & Lb_Theorem(
         Current_Language) &
         To_Unbounded_String("2). ") & Unbounded_Lf;

   end Dispatched_Bound_On_Processor_Utilization;



   procedure Dispatched_Response_Time (
         My_Scheduler   : in     Llf_Scheduler;    
         My_Tasks       : in out Tasks_Set;        
         Processor_Name : in     Unbounded_String; 
         Msg            : in out Unbounded_String; 
         Response_Time  :    out Response_Table    ) is 
   begin

      null;
   end Dispatched_Response_Time;



   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Llf_Scheduler;          
         Si             : in out Scheduling_Information; 
         Processor_Name : in     Unbounded_String;       
         My_Tasks       : in out Tasks_Set;              
         My_Resources   : in out Resources_Set           ) is 


   begin
      null;
   end Specific_Scheduler_Initialization;







end Scheduler.Llf ;


