------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Text_Io;
use Text_Io;

with Tasks;
use Tasks;
with Resources;
use Resources;

with Natural_Util;
use Natural_Util;

with Config;
use Config;



package body Scheduler is




   procedure Set_Preemptive (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Preempt      : in     Preemptive_Type          ) is 
   begin
      My_Scheduler.Is_Preemptive:=Preempt;
   end Set_Preemptive;


   function Get_Preemptive (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Preemptive_Type is 
   begin
      return My_Scheduler.Is_Preemptive;
   end Get_Preemptive;


   procedure Dump (
         My_Scheduler : in     Generic_Scheduler'Class ) is 
   begin
      Put_Line("Dump scheduler :");
      Put_Line("Scheduler name : " & My_Scheduler.Name'Img);
      Put_Line("Is preemptive  : " & My_Scheduler.Is_Preemptive'Img);
      Put_Line("Quantum        : " & My_Scheduler.Quantum'Img);
   end Dump;


   procedure Initialize (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      Reset(A_Scheduler);
   end Initialize;


   procedure Finalize (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      null;
   end Finalize;


   procedure Adjust (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      null;
   end Adjust;



   procedure Reset (
         A_Scheduler : in out Generic_Scheduler'Class ) is 
   begin
      A_Scheduler.Is_Preemptive:=Preemptive;
      A_Scheduler.Quantum:=0;
   end Reset;



   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String is 
   begin
      return To_Unbounded_String(My_Scheduler.Name'Img);
   end Get_Name;

   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Scheduler_Type is 
   begin
      return My_Scheduler.Name;
   end Get_Name;



   function Scheduling_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Natural is 

      Period : Natural := 1;  

      A_Task      : Task_Ptr;  
      My_Iterator : Tasks_Iterator;  

   begin


      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks, A_Task, My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Period:= Natural_Util.Lcm(Period,
               Periodic_Task_Ptr(A_Task).Period);
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);
      end loop;


      return Period;

   end Scheduling_Period;


   function Processor_Utilization_Over_Deadline (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(A_Task.Deadline));
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;


      return Utilization;

   end Processor_Utilization_Over_Deadline;


   function Processor_Utilization_Over_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);



      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(Periodic_Task_Ptr(A_Task).Period));
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Utilization;

   end Processor_Utilization_Over_Period;




   procedure Scheduler_Initialization (
         My_Scheduler      : in out Generic_Scheduler'Class; 
         Si                : in out Scheduling_Information;  
         Processor_Name    : in     Unbounded_String;        
         My_Tasks          : in out Tasks_Set;               
         My_Resources      : in out Resources_Set;           
         With_Offsets      :        Boolean;                 
         With_Precedencies :        Boolean;                 
         With_Resources    :        Boolean                  ) is 

      Iterator1  : Tasks_Iterator;  
      Iterator2  : Resources_Iterator;  
      A_Task     : Task_Ptr;  
      A_Resource : Resource_Ptr;  

   begin

      My_Scheduler.Previously_Elected := Tasks_Range'First;


      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, A_Task, Iterator1);

         Si.Tcbs(Si.Number_Of_Tasks):=Build_Tcb(
            My_Scheduler, A_Task);

         Si.Number_Of_Tasks:=Si.Number_Of_Tasks+1;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);


      end loop;


      if (With_Resources=True)
            then
         if (Get_Number_Of_Elements(My_Resources)>0)
               then

            Reset_Iterator(Iterator2);
            loop

               Current_Element(My_Resources, A_Resource, Iterator2);

               Si.Shared_Resources(Si.
                  Number_Of_Resources):=Build_Resource(My_Scheduler,
                  A_Resource);

               Si.Number_Of_Resources:=Si.
                  Number_Of_Resources+1;

               exit when Is_Last_Element(My_Resources, Iterator2) = True;
               Next_Element(My_Resources, Iterator2);


            end loop;
         end if;
      end if;



   end Scheduler_Initialization;



   procedure Build_Scheduling_Sequence (
         My_Scheduler      : in out Generic_Scheduler'Class;                    
         My_Tasks          : in     Tasks_Set;                                  
         My_Dependencies   : in     Tasks_Dependencies;                         
         Processor_Name    : in     Unbounded_String;                           
         Result            :    out Scheduling_Sequence;                        
         Given_Last_Time   :        Natural                 := 0;               
         My_Resources      : in     Resources_Set           := No_Resource_Set; 
         With_Offsets      :        Boolean                 := True;            
         With_Precedencies :        Boolean                 := True;            
         With_Resources    :        Boolean                 := True             ) is 



      Tmp_Tasks     : Tasks_Set;  
      Tmp_Resources : Resources_Set;  

      Last_Time : Natural;  

      Current_Time      : Natural         := 0;  
      Current_Time_Unit : Time_Unit_Range := 0;  
      Elected           : Tasks_Range     := Tasks_Range'First;  
      No_Task           : Boolean;  

      Si : Scheduling_Information;  

   begin


      Si.Number_Of_Tasks:=0;
      Si.Number_Of_Resources:=0;




      -- Before scheduling, check if th tasks can be
      -- schedule with the selected scheduler
      --

      Check_Before_Scheduling(My_Scheduler, My_Tasks, Processor_Name);



      -- First : gather information
      --
      Current_Processor_Name:=Processor_Name;

      Select_And_Copy(
         My_Tasks, Tmp_Tasks, Select_Cpu'access);



      if (With_Resources=True)
            then
         if (Get_Number_Of_Elements(My_Resources)>0)
               then
            Select_And_Copy(
               My_Resources, Tmp_Resources, Select_Cpu'access);
         end if;
      end if;


      Scheduler_Initialization (
         My_Scheduler, Si, Processor_Name, Tmp_Tasks, Tmp_Resources,
         With_Offsets, With_Precedencies, With_Resources);


      Specific_Scheduler_Initialization(
         My_Scheduler, Si, Processor_Name, Tmp_Tasks, Tmp_Resources);




      -- Compute last scheduled time unit
      --

      if(Given_Last_Time = 0)
            then
         Last_Time:=Natural'Min (Max_Scheduling_Period,
            Scheduling_Period (My_Scheduler, Tmp_Tasks, Processor_Name));
      else
         Last_Time:=Given_Last_Time;
      end if;



      -- Second : compute scheduling
      --

      Reset(Result);

      while(Current_Time < Last_Time) loop


         Do_Election(My_Scheduler, Si, Processor_Name, Current_Time,
            My_Dependencies,
            With_Offsets, With_Precedencies, With_Resources,
            Elected, No_Task);


         -- No task to schedule
         --
         if (No_Task = True)
               then
            Current_Time:=Current_Time+1;

         else
            Run_Task(My_Scheduler, Si, Elected, Result,
               Current_Time, Current_Time_Unit,
               With_Offsets, With_Precedencies, With_Resources);

         end if;

      end loop;

   end Build_Scheduling_Sequence;




   procedure Run_Task (
         My_Scheduler      : in out Generic_Scheduler;              
         Si                : in out Scheduling_Information;         
         Elected           : in     Tasks_Range;                    
         Result            : in out Scheduling_Sequence;            
         Current_Time      : in out Natural;                        
         Current_Time_Unit : in out Time_Unit_Range;                
         With_Offsets      :        Boolean                := True; 
         With_Precedencies :        Boolean                := True; 
         With_Resources    :        Boolean                := True  ) is 


   begin

      -- Allocate/release shared resources
      --
      if(With_Resources=True)
            then
         Allocate_Resource(My_Scheduler, Si,
            Si.Tcbs(Elected));

         Release_Resource(My_Scheduler, Si,
            Si.Tcbs(Elected));
      end if;


      if (My_Scheduler.Is_Preemptive = Preemptive)
            then
         Result.Scheduled.Entries(Current_Time_Unit).Data:=
            Current_Time;
         Result.Scheduled.Entries(Current_Time_Unit).Item:=
            Si.Tcbs(
            Elected).Tsk;
         Result.Scheduled.Nb_Entries:=Result.Scheduled.Nb_Entries+ 1;
         Si.Tcbs(Elected).Rest_Of_Capacity:=Si.
            Tcbs(Elected).
            Rest_Of_Capacity - 1;


         if (Current_Time_Unit > 0)
               then

            if ( (Result.Scheduled.Entries(Current_Time_Unit - 1).
                     Item.
                     Name /=
                     Result.Scheduled.Entries(Current_Time_Unit).
                     Item.
                     Name)
                  and (Result.Scheduled.Entries(Current_Time_Unit -
                        1).Data+
                     1=
                     Result.Scheduled.Entries(Current_Time_Unit).
                     Data)
                  and (Si.Tcbs(My_Scheduler.Previously_Elected).
                     Rest_Of_Capacity/=0)
                  and (Si.Tcbs(My_Scheduler.Previously_Elected).
                     Rest_Of_Capacity/=
                     Si.Tcbs(My_Scheduler.Previously_Elected).Tsk.
                     Capacity)
                  )
                  then
               Result.Number_Of_Preemption:= Result.
                  Number_Of_Preemption + 1;
            end if;
         end if;

         Current_Time_Unit:=Current_Time_Unit + 1;
         Current_Time:=Current_Time + 1;

      else
         for I in 0 .. (Si.Tcbs(Elected).Tsk.Capacity - 1)
               loop
            Result.Scheduled.Entries(Current_Time_Unit).Data:=
               Current_Time;
            Result.Scheduled.Entries(Current_Time_Unit).Item:=
               Si.Tcbs(
               Elected).Tsk;
            Current_Time_Unit:=Current_Time_Unit + 1;
            Result.Scheduled.Nb_Entries:=Result.Scheduled.
               Nb_Entries+
               1;
            Current_Time:=Current_Time + 1;
         end loop;
         Si.Tcbs(Elected).Rest_Of_Capacity:=0;
      end if;


      -- Next activation ?
      --
      if (Si.Tcbs(Elected).Rest_Of_Capacity = 0)
            then
         Si.Tcbs(Elected).End_Time:=Current_Time;
         if (Si.Tcbs(Elected).Tsk.Task_Type=Periodic_Type)
               then
            Si.Tcbs(Elected).Activation:=Si.
               Tcbs(Elected).Activation+1;
            Si.Tcbs(Elected).Rest_Of_Capacity:=
               Si.Tcbs(Elected).Tsk.
               Capacity;
            Si.Tcbs(Elected).Wake_Up_Time:=Si.
               Tcbs(Elected).Wake_Up_Time +
               Periodic_Task_Ptr(Si.Tcbs(Elected).Tsk).
               Period;
         end if;
      end if;


      My_Scheduler.Previously_Elected:=Elected;


   end Run_Task;



   procedure Compute_Response_Time (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in out Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Msg            : in out Unbounded_String;        
         Response_Time  :    out Response_Table           ) is 
   begin
      Double_Tasks_Parameters.Initialize(Response_Time);
      Dispatched_Response_Time(My_Scheduler, My_Tasks, Processor_Name, Msg,
         Response_Time);
   end Compute_Response_Time;


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Result         : in out Double;                  
         Msg            : in out Unbounded_String         ) is 
   begin
      Dispatched_Bound_On_Processor_Utilization(My_Scheduler,My_Tasks,
         Processor_Name, Result, Msg);
   end Bound_On_Processor_Utilization;


   -- Return true is the task "task_id" can be schedule now
   -- according its offsets
   --		
   function Check_Offset (
         A_Tcb   : Tcb_Ptr; 
         At_Time : Natural  ) 
     return Boolean is 
      Target_Time : Integer := 0;  
   begin

      for K in Offset_Range loop
         if A_Tcb.Tsk.Offset(K).Activation=A_Tcb.Activation
               then
            Target_Time:=A_Tcb.Wake_Up_Time+A_Tcb.Tsk.Offset(
               K).Value;
            if At_Time<Target_Time
                  then
               return False;
            end if;
         end if;
      end loop;

      return True;
   end Check_Offset;





   function Select_Cpu (
         Op1 : in     Task_Ptr ) 
     return Boolean is 
   begin
      return (Op1.Location = Current_Processor_Name);
   end Select_Cpu;


   function Select_Cpu (
         Op1 : in     Resource_Ptr ) 
     return Boolean is 
   begin
      return (Op1.Location = Current_Processor_Name);
   end Select_Cpu;





   procedure Release_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        :        Tcb_Ptr                 ) is 

      Current_Time : Natural := 0;  

   begin

      if (Si.Number_Of_Resources>0)
            then
         Current_Time:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0.. Si.Number_Of_Resources -1 loop


            -- For each resource, check if the task hold resources
            -- and then release them
            --
            for Index2 in 0 .. Si.Shared_Resources(Index1).
                  Shared.Task_Tab.Nb_Entries-1
                  loop


               if (Current_Time=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_End)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then
                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State+1;
                  for I in 0 .. Si.Shared_Resources(Index1).
                        Nb_Allocated-1 loop
                     if Si.Shared_Resources(Index1).
                           Allocated_By(I)=A_Tcb.Tsk.Name
                           then
                        Si.Shared_Resources(Index1).
                           Allocated_By(I):= Si.
                           Shared_Resources(Index1).Allocated_By(
                           Si.Shared_Resources(Index1).
                           Nb_Allocated-1);
                        Si.Shared_Resources(Index1).
                           Nb_Allocated:=Si.Shared_Resources(
                           Index1).Nb_Allocated-1;
                        exit;
                     end if;
                  end loop;

               end if;


            end loop;


         end loop;
      end if;

   end Release_Resource;



   procedure Allocate_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        :        Tcb_Ptr                 ) is 

      Current_Time : Natural := 0;  


   begin

      if (Si.Number_Of_Resources>0)
            then

         Current_Time:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0 .. Si.Number_Of_Resources -1 loop


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            -- otherwise, block the task
            --
            for Index2 in 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  loop


               if (Current_Time=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then


                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State-1;

                  Si.Shared_Resources(Index1).Allocated_By(
                     Si.Shared_Resources(Index1).Nb_Allocated):=
                     A_Tcb.Tsk.Name;
                  Si.Shared_Resources(Index1).Nb_Allocated:=
                     Si.Shared_Resources(Index1).Nb_Allocated+1;
               end if;

            end loop;


         end loop;
      end if;


   end Allocate_Resource;


   procedure Check_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        : in     Tcb_Ptr;                
         Is_Ready     :    out Boolean                 ) is 


      Current_Time : Natural := 0;  

   begin

      Is_Ready:=True;

      if (Si.Number_Of_Resources>0)
            then

         Current_Time:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         for Index1 in 0 .. Si.Number_Of_Resources -1 loop


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            --
            for Index2 in 0 ..Si.Shared_Resources(Index1).
                  Shared.Task_Tab.Nb_Entries-1
                  loop


               if (Current_Time=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     and (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     then


                  if (Si.Shared_Resources(Index1).Shared.State<=
                        0)
                        then
                     Is_Ready:=False;
                  end if;
               end if;


            end loop;


         end loop;
      end if;


   end Check_Resource;




   procedure Set_Quantum (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Q            : in     Natural                  ) is 
   begin
      My_Scheduler.Quantum:=Q;
   end Set_Quantum;


   function Get_Quantum (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Natural is 
   begin
      return My_Scheduler.Quantum;
   end Get_Quantum;


   function Get_Quantum (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String is 
   begin
      return To_Unbounded_String(My_Scheduler.Quantum'Img);
   end Get_Quantum;




   function Build_Resource (
         My_Scheduler : in     Generic_Scheduler; 
         A_Resource   :        Resource_Ptr       ) 
     return Shared_Resource_Ptr is 

      New_A_Resource : Shared_Resource_Ptr;  

   begin

      New_A_Resource:= new Shared_Resource;
      New_A_Resource.Shared:=A_Resource;

      return New_A_Resource;

   end Build_Resource;




   procedure Initialize (
         S : in out Scheduling_Information ) is 
   begin
      S.Number_Of_Tasks  := 0;
      S.Number_Of_Resources := 0;
   end Initialize;


   function Check_Precedencies (
         Si           : in     Scheduling_Information; 
         Deps         :        Tasks_Dependencies;     
         Current_Time : in     Natural;                
         A_Tcb        :        Tcb_Ptr                 ) 
     return Boolean is 


      Previous : Tasks_Set;  
      Left     : Task_Ptr;  
      Ite1     : Tasks_Iterator;  


   begin

      if Has_Predecessor(Deps, A_Tcb.Tsk) then

         Reset(Previous);
         Previous:=Get_Predecessors_List(Deps, A_Tcb.Tsk);

         Reset_Iterator(Ite1);

         for J in 0 .. Get_Number_Of_Elements(Previous)-1 loop

            Current_Element(Previous, Left, Ite1);

            for I in 0..Si.Number_Of_Tasks-1 loop
               if Si.Tcbs(I).Tsk.Name = Left.Name
                     then


                  if (Si.Tcbs(I).Activation<=
                        A_Tcb.Activation)
                        then
                     return False;
                  end if;

                  -- Same activation ... be sure the previous task
                  -- is ended
                  --

                  if (Si.Tcbs(I).Activation=A_Tcb.Activation+1)
                        and (Si.Tcbs(I).End_Time>Current_Time)
                        then
                     return False;
                  end if;

               end if;
            end loop;


            Next_Element(Previous, Ite1);
         end loop;
      end if;


      return True;

   end Check_Precedencies;



end Scheduler;


