------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Tasks;
use Tasks;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Resources;
use Resources;
use Resources.Task_Lists;

with Resource_Set;
use Resource_Set;
use Resource_Set.Generic_Resource_Set;

with Ada.Finalization;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Config;
use Config;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Scheduling_Sequences;
use Scheduling_Sequences;

with Dependencies;
use Dependencies;
use Dependencies.Dep_Set;
with Task_Dependencies;
use Task_Dependencies;


with Tables;

with Double_Util;

package Scheduler is



   type Scheduler_Type is 
         (Earliest_Deadline_First, 
          Least_Laxity_First,      
          Rate_Monotonic,          
          Deadline_Monotonic,      
          Highiest_Priority_First); 

   type Preemptive_Type is 
         (Preemptive,    
          Non_Preemptive); 



   package Double_Tasks_Parameters is new Tables(Double, Task_Ptr,
      Config.Max_Tasks, 0, Double_Util.Put, Double_Util.Initialize, Tasks.
      Put_Name);
   subtype Response_Table is Double_Tasks_Parameters.Table;
   subtype Response_Range is Double_Tasks_Parameters.Table_Range;


   No_Resource_Set : Resources_Set;  

   Processor_Utilization_Exceeded : exception;  

   ----------------------------------------
   type Tcb is tagged 
      record 
         Tsk              : Task_Ptr;  
         Wake_Up_Time     : Natural;  
         Rest_Of_Capacity : Natural;  
         Activation       : Natural;  
         End_Time         : Natural;  
      end record; 

   type Tcb_Ptr is access all Tcb'Class; 
   type Tcb_Table is array (Tasks_Range) of Tcb_Ptr; 

   ----------------------------------------

   type Allocation_Table is array (Natural range 0 .. Max_Tasks_For_A_Resource) of Unbounded_String; 

   type Shared_Resource is tagged 
      record 
         Shared       : Resource_Ptr;  
         Allocated_By : Allocation_Table;  
         Nb_Allocated : Natural          := 0;  
      end record; 

   type Shared_Resource_Ptr is access all Shared_Resource'Class; 
   type Shared_Resource_Table is array (Resources_Range) of Shared_Resource_Ptr; 

   ----------------------------------------

   type Scheduling_Information is 
      record 
         Tcbs            : Tcb_Table;  
         Number_Of_Tasks : Tasks_Range := 0;  

         Shared_Resources    : Shared_Resource_Table;  
         Number_Of_Resources : Resources_Range       := 0;  
      end record; 


   procedure Initialize (
         S : in out Scheduling_Information ); 

   ----------------------------------------

   type Generic_Scheduler is abstract tagged private; 
   type Scheduler_Ptr is access all Generic_Scheduler'Class; 


   procedure Initialize (
         A_Scheduler : in out Generic_Scheduler ); 


   procedure Finalize (
         A_Scheduler : in out Generic_Scheduler ); 


   procedure Adjust (
         A_Scheduler : in out Generic_Scheduler ); 




   procedure Reset (
         A_Scheduler : in out Generic_Scheduler'Class ); 


   procedure Build_Scheduling_Sequence (
         My_Scheduler      : in out Generic_Scheduler'Class;                    
         My_Tasks          : in     Tasks_Set;                                  
         My_Dependencies   : in     Tasks_Dependencies;                         
         Processor_Name    : in     Unbounded_String;                           
         Result            :    out Scheduling_Sequence;                        
         Given_Last_Time   :        Natural                 := 0;               
         My_Resources      : in     Resources_Set           := No_Resource_Set; 
         With_Offsets      :        Boolean                 := True;            
         With_Precedencies :        Boolean                 := True;            
         With_Resources    :        Boolean                 := True             ); 



   procedure Compute_Response_Time (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in out Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Msg            : in out Unbounded_String;        
         Response_Time  :    out Response_Table           ); 


   procedure Set_Preemptive (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Preempt      : in     Preemptive_Type          ); 


   function Get_Preemptive (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Preemptive_Type; 


   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Scheduler_Type; 

   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String; 


   procedure Set_Quantum (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Q            : in     Natural                  ); 

   function Get_Quantum (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Natural; 

   function Get_Quantum (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String; 


   function Scheduling_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Natural; 

   function Processor_Utilization_Over_Deadline (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double; 


   function Processor_Utilization_Over_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double; 


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Result         : in out Double;                  
         Msg            : in out Unbounded_String         ); 


   procedure Dispatched_Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in     Tasks_Set;         
         Processor_Name : in     Unbounded_String;  
         Result         : in out Double;            
         Msg            : in out Unbounded_String   ) is 
   abstract;


   procedure Dispatched_Response_Time (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in out Tasks_Set;         
         Processor_Name : in     Unbounded_String;  
         Msg            : in out Unbounded_String;  
         Response_Time  :    out Response_Table     ) is 
   abstract;


   procedure Dump (
         My_Scheduler : in     Generic_Scheduler'Class ); 

   --
   --  Procedure and functions for scheduling
   --

   procedure Run_Task (
         My_Scheduler      : in out Generic_Scheduler;              
         Si                : in out Scheduling_Information;         
         Elected           : in     Tasks_Range;                    
         Result            : in out Scheduling_Sequence;            
         Current_Time      : in out Natural;                        
         Current_Time_Unit : in out Time_Unit_Range;                
         With_Offsets      :        Boolean                := True; 
         With_Precedencies :        Boolean                := True; 
         With_Resources    :        Boolean                := True  ); 


   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in     Tasks_Set;         
         Processor_Name : in     Unbounded_String   ) is 
   abstract;

   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Generic_Scheduler;      
         Si             : in out Scheduling_Information; 
         Processor_Name : in     Unbounded_String;       
         My_Tasks       : in out Tasks_Set;              
         My_Resources   : in out Resources_Set           ) is 
   abstract;

   procedure Scheduler_Initialization (
         My_Scheduler      : in out Generic_Scheduler'Class; 
         Si                : in out Scheduling_Information;  
         Processor_Name    : in     Unbounded_String;        
         My_Tasks          : in out Tasks_Set;               
         My_Resources      : in out Resources_Set;           
         With_Offsets      :        Boolean;                 
         With_Precedencies :        Boolean;                 
         With_Resources    :        Boolean                  ); 


   procedure Do_Election (
         My_Scheduler      : in out Generic_Scheduler;      
         Si                : in out Scheduling_Information; 
         Processor_Name    : in     Unbounded_String;       
         At_Time           :        Natural;                
         My_Dependencies   : in     Tasks_Dependencies;     
         With_Offsets      :        Boolean;                
         With_Precedencies :        Boolean;                
         With_Resources    :        Boolean;                
         Elected           : in out Tasks_Range;            
         No_Task           : in out Boolean                 ) is 
   abstract;


   function Build_Tcb (
         My_Scheduler : in     Generic_Scheduler; 
         A_Task       :        Task_Ptr           ) 
     return Tcb_Ptr is 
   abstract;


   function Build_Resource (
         My_Scheduler : in     Generic_Scheduler; 
         A_Resource   :        Resource_Ptr       ) 
     return Shared_Resource_Ptr; 


   procedure Allocate_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        :        Tcb_Ptr                 ); 

   procedure Release_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        :        Tcb_Ptr                 ); 

   procedure Check_Resource (
         My_Scheduler : in     Generic_Scheduler;      
         Si           : in out Scheduling_Information; 
         A_Tcb        : in     Tcb_Ptr;                
         Is_Ready     :    out Boolean                 ); 



   function Check_Offset (
         A_Tcb   : Tcb_Ptr; 
         At_Time : Natural  ) 
     return Boolean; 


   function Check_Precedencies (
         Si           : in     Scheduling_Information; 
         Deps         :        Tasks_Dependencies;     
         Current_Time : in     Natural;                
         A_Tcb        :        Tcb_Ptr                 ) 
     return Boolean; 





   Current_Processor_Name : Unbounded_String;  

   function Select_Cpu (
         Op1 : in     Task_Ptr ) 
     return Boolean; 
   function Select_Cpu (
         Op1 : in     Resource_Ptr ) 
     return Boolean; 





private

   type Generic_Scheduler is abstract new Ada.Finalization.Controlled with 
      record 
         Name          : Scheduler_Type;  
         Is_Preemptive : Preemptive_Type;  
         Quantum       : Natural         := 0;  

         Previously_Elected : Tasks_Range := Tasks_Range'First;  
      end record; 




end Scheduler ;


