------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Fixed_Strings;
use Fixed_Strings;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Config;
use Config;

with Tasks;
use Tasks;

with Offsets;
use Offsets;

with Scheduling_Sequences;
use Scheduling_Sequences;


with Sets;


package Task_Set is

   Task_Must_Be_Periodic                   : exception;  
   Task_Model_Error                        : exception;  
   Task_Not_Found                          : exception;  
   Task_Already_Defined                    : exception;  
   Invalid_Parameter                       : exception;  
   Task_Must_Have_Period_Equal_To_Deadline : exception;  
   Processor_Utilization_Exceeded          : exception;  
   Priority_Error                          : exception;  
   Start_Time_Error                        : exception;  
   Offset_Error                            : exception;  


   type Task_Descriptor is 
      record 
         Name          : Fixed_String;  
         Location      : Fixed_String;  
         Capacity      : Natural;  
         Start_Time    : Natural;  
         Priority      : Priority_Range;  
         Policy        : Policies;  
         Period        : Natural;  
         Blocking_Time : Natural;  
         Deadline      : Natural;  
         Jitter        : Natural;  
         Offset        : Offset_Table;  
      end record; 


   type Task_Parameters is 
         (Location,      
          Capacity,      
          Start_Time,    
          Priority,      
          Period,        
          Blocking_Time, 
          Deadline,      
          Jitter); 



   package Generic_Task_Set is new Sets (
      Max_Element     => Config.Max_Tasks, 
      File_Descriptor => Task_Descriptor,  
      Element         => Task_Ptr,         
      Copy            => Copy,             
      Free            => Free,             
      Put             => Dump);


   use Generic_Task_Set;

   type Tasks_Set is new Generic_Task_Set.Set with private; 


   subtype Tasks_Range is Generic_Task_Set.Element_Range;
   subtype Tasks_Iterator is   Generic_Task_Set.Iterator;




   procedure Fill_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     :    out Task_Descriptor; 
         A_Task   : in     Task_Ptr         ); 


   procedure Extract_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     : in     Task_Descriptor; 
         A_Task   :    out Task_Ptr         ); 


   procedure Add_Task (
         My_Tasks      : in out Tasks_Set;        
         Name          : in     Unbounded_String; 
         Location      : in     Unbounded_String; 
         Start_Time    : in     Natural;          
         Capacity      : in     Natural;          
         Period        : in     Natural;          
         Deadline      : in     Natural;          
         Jitter        : in     Natural;          
         Blocking_Time : in     Natural;          
         Priority      : in     Priority_Range;   
         Policy        : in     Policies;         
         Offset        :        Offset_Table      ); 

   procedure Add_Task (
         My_Tasks      : in out Tasks_Set;        
         A_Task        : in out Task_Ptr;         
         Name          : in     Unbounded_String; 
         Location      : in     Unbounded_String; 
         Start_Time    : in     Natural;          
         Capacity      : in     Natural;          
         Period        : in     Natural;          
         Deadline      : in     Natural;          
         Jitter        : in     Natural;          
         Blocking_Time : in     Natural;          
         Priority      : in     Priority_Range;   
         Policy        : in     Policies;         
         Offset        :        Offset_Table      ); 

   procedure Check_Temporal_Constraints (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Unbounded_String     ); 

   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Unbounded_String; 

   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Priority_Range; 

   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Natural; 

   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Unbounded_String  ); 

   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Priority_Range    ); 

   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Natural           ); 



   function Get_Number_Of_Task_From_Processor (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Tasks_Range; 


   function Search_Task (
         My_Tasks : in     Tasks_Set;       
         Name     : in     Unbounded_String ) 
     return Task_Ptr; 


   procedure Delete_Processor (
         My_Tasks    : in out Tasks_Set;       
         A_Processor : in     Unbounded_String ); 


   function Increasing_Name (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Increasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Decreasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Increasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Decreasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Increasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 

   function Decreasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean; 


   function Deadline_Inferior_To_Period (
         My_Tasks : in     Tasks_Set ) 
     return Boolean; 

   function Is_Harmonic (
         My_Tasks : in     Tasks_Set ) 
     return Boolean; 


   procedure Have_Deadlines_Equal_Than_Periods_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 
   procedure Periodic_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 
   procedure Priority_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 
   procedure Start_Time_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 
   procedure Offset_Control (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ); 

private

   type Tasks_Set is new Generic_Task_Set.Set with null record; 




end Task_Set;



