------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Project;
use Project;

with Resources;
use Resources;
with Resource_Set;
use Resource_Set;
use Resource_Set.Generic_Resource_Set;

with Processors;
use Processors;
with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Unbounded_Strings;
use Unbounded_Strings;

with Config;
use Config;


with Scheduler;
use Scheduler;
with Scheduler.Fixed_Priority;
use Scheduler.Fixed_Priority;

with Blocking_Time;
use Blocking_Time;


package body Call_Blocking_Time_Logic is


   procedure Display_Blocking_Time_For_All_Tasks (
         A_Task_Set           : in     Tasks_Set;          
         A_Window             : in     Root_Window_Access; 
         Inject_Blocking_Time :        Boolean             ) is 



      My_Task_Iterator1     : Tasks_Iterator;  
      My_Task_Iterator2     : Tasks_Iterator;  
      My_Resource_Iterator  : Resources_Iterator;  
      My_Processor_Iterator : Processors_Iterator;  

      Tmp         : Tasks_Set;  
      A_Task      : Task_Ptr;  
      A_Task_Tmp  : Task_Ptr;  
      A_Resource  : Resource_Ptr;  
      A_Processor : Processor_Ptr;  

   begin



      Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf & Unbounded_Lf);

      Root_Window_Pkg.Write_Title(A_Window, Lb_Blocking_Time(
            Current_Language) & Lb_Colon &  To_Unbounded_String(" (") &
         Lb_See(Current_Language) & To_Unbounded_String("[3]).") &
         Unbounded_Lf  );



      -- Check protocol according to the scheduler
      -- If PIP/PCP resource, then scheduler 
      -- can only be RM, DM or HPF
      --
      Reset_Iterator(My_Resource_Iterator);

      loop
         Current_Element(Sys.Resources,A_Resource,My_Resource_Iterator);


         if (A_Resource.Protocol =
               Priority_Ceiling_Protocol)
               or  (A_Resource.Protocol =
               Priority_Inheritance_Protocol)
               then


            Reset_Iterator(My_Processor_Iterator);

            loop

               Current_Element(Sys.Processors, A_Processor,
                  My_Processor_Iterator);

               if (A_Processor.Name = A_Resource.Location)
                     then
                  if
                        (Get_Name(A_Processor.Scheduler.all) /=
                        Rate_Monotonic) and
                        (Get_Name(A_Processor.Scheduler.all) /=
                        Deadline_Monotonic) and
                        (Get_Name(A_Processor.Scheduler.all) /=
                        Highiest_Priority_First)
                        then

                     Root_Window_Pkg.Write_Text(A_Window, Unbounded_Lf &
                        Lb_Warning(Current_Language) & A_Resource.Name &
                        Lb_Compute_Blocking_Error2(Current_Language) &
                        Unbounded_Lf);
                  end if;
               end if;


               exit when Is_Last_Element(Sys.Processors,
                  My_Processor_Iterator) = True;

               Next_Element(Sys.Processors, My_Processor_Iterator);

            end loop;
         end if;



         exit when Is_Last_Element(Sys.Resources, My_Resource_Iterator) =
            True;

         Next_Element(Sys.Resources, My_Resource_Iterator);

      end loop;



      -- Copy the task set before modification
      --
      Duplicate(A_Task_Set, Tmp);


      -- Check if scheduler is RM/DM/HPF
      -- and then, change task_set priorities
      --
      Reset_Iterator(My_Processor_Iterator);

      loop

         Current_Element(Sys.Processors, A_Processor,
            My_Processor_Iterator);

         if (Get_Name(A_Processor.Scheduler.all) =
               Rate_Monotonic)
               then
            Set_Priority_According_To_Rm(
               Fixed_Priority_Scheduler(A_Processor.Scheduler.all),
               Tmp,
               A_Processor.Name);
         else
            if (Get_Name(A_Processor.Scheduler.all) =
                  Deadline_Monotonic)
                  then
               Set_Priority_According_To_Dm(
                  Fixed_Priority_Scheduler(A_Processor.Scheduler.all),
                  Tmp,
                  A_Processor.Name);
            end if;
         end if;

         exit when Is_Last_Element(Sys.Processors, My_Processor_Iterator) =
            True;
         Next_Element(Sys.Processors, My_Processor_Iterator);

      end loop;


      -- And now, compute blocking time
      --

      Reset_Iterator(My_Task_Iterator1);

      loop

         Current_Element(Tmp, A_Task, My_Task_Iterator1);

         A_Task.Blocking_Time :=
            Compute_Blocking_Time_Task(Tmp, Sys.Resources, A_Task.Name);

         Root_Window_Pkg.Write_Text(A_Window, Unbounded_Lf & A_Task.Name &
            " => " & A_Task.Blocking_Time'Img );

         exit when Is_Last_Element(Tmp, My_Task_Iterator1) = True;

         Next_Element(Tmp, My_Task_Iterator1);

      end loop;




      -- Update blocking time
      --
      if (Inject_Blocking_Time=True)
            then

         Reset_Iterator(My_Task_Iterator1);
         loop
            Current_Element(Tmp, A_Task_Tmp, My_Task_Iterator1);

            Reset_Iterator(My_Task_Iterator2);
            loop
               Current_Element(Sys.Tasks, A_Task, My_Task_Iterator2);

               if (A_Task.Name = A_Task_Tmp.Name)
                     then
                  A_Task.Blocking_Time :=
                     A_Task_Tmp.Blocking_Time;
               end if;

               exit when Is_Last_Element(Sys.Tasks, My_Task_Iterator2) =
                  True;
               Next_Element(Sys.Tasks, My_Task_Iterator2);
            end loop;


            exit when Is_Last_Element(Tmp, My_Task_Iterator1) = True;
            Next_Element(Tmp, My_Task_Iterator1);

         end loop;


         Root_Window_Pkg.Write_Text(A_Window,Unbounded_Lf & Unbounded_Lf &
            Lb_Inject_Success(Current_Language) & Unbounded_Lf);
      end if;


   exception
      when Tasks.Compute_Aborded =>
         begin
            Write_Text(A_Window, Unbounded_Lf & Lb_Compute_Blocking_Error1(
                  Current_Language) & Unbounded_Lf);
         end;

   end Display_Blocking_Time_For_All_Tasks;


end Call_Blocking_Time_Logic;









