------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Config;
use Config;
with Project;
use Project;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;
with Scheduler.Fixed_Priority;
use  Scheduler.Fixed_Priority;
with Scheduler.Fixed_Priority.Dm;
use  Scheduler.Fixed_Priority.Dm;
with Scheduler.Fixed_Priority.Rm;
use  Scheduler.Fixed_Priority.Rm;

with Double_Util;
use Double_Util;
with Translate;
use Translate;
with Tasks;
use Tasks;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;
with Unbounded_Strings;
use Unbounded_Strings;


with Task_Dependencies;
use  Task_Dependencies;
with Dependencies_Services;
use Dependencies_Services;

with Multiprocessor_Services;
use Multiprocessor_Services;
use Multiprocessor_Services.Scheduling_Result_Per_Processor;



package body Call_Scheduling_Logic  is




   procedure Build_And_Check_Scheduling (
         A_Window : in     Root_Window_Access; 
         Period   : in     Natural             ) is 


      Check   : Unbounded_String := Empty_String;  
      Message : Unbounded_String;  

   begin

      Sort(Sys.Tasks, Increasing_Name'access);
      Scheduling_Base_Period:=Period;


      Build_Multiprocessor_Scheduling(
         Sys.Processors, Sys.Tasks, Sys.Resources, Sys.Dependencies, Sched,
         Period,
         Schedule_With_Offsets, Schedule_With_Precedencies,
         Schedule_With_Resources);


      for J in 0.. Sched.Nb_Entries-1 loop


         Write_Title(Root_Window, Lb_Processor(Current_Language) &
            Sched.Entries(J).Data.Processor.Name  & To_Unbounded_String(
               " : ") &
            Unbounded_Lf & Unbounded_Lf);


         if Sched.Entries(J).Data.Msg/=Empty_String then
            Write_Text(A_Window, Sched.Entries(J).Data.Msg);

         else

            Message:= Lb_Number_Of_Preemption(Current_Language) & Sched.
               Entries(J).Data.Result.
               Number_Of_Preemption'Img & Unbounded_Lf;

            Check_Temporal_Constraints (Sys.Tasks, Sched.Entries(J).Data.
               Result, Sched.Entries(J).Data.Processor.Name,
               Check);
            if Check = Empty_String
                  then
               Message:=Message & Lb_No_Missed_Deadline(Current_Language) &
                  Unbounded_Lf;
            else
               Message:=Message & Check & Unbounded_Lf;
            end if;

            Write_Text(A_Window, Message);

         end if;

         Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf);

      end loop;


   end Build_And_Check_Scheduling;






   procedure Compute_Scheduling_Period (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 
      Period  : Natural;  
      Result  : Double;  
      Free_Ut : Double;  
   begin
      Period:=Scheduling_Period(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name);
      Message:=Message & Lb_Scheduling_Period(Current_Language) & Period'
         Img  &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & Unbounded_Lf;

      Result:=Processor_Utilization_Over_Period(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);

      Free_Ut:=Double(Period)*(1.0-Result);
      Free_Ut:=Double'Max(0.0,Free_Ut);
      Message:=Message & Lb_Minus
         & Format(Free_Ut) &
         Lb_Free_Unit_Time(Current_Language) & Unbounded_Lf;


   exception
      when Task_Set.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_2(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_3(Current_Language) &
            Unbounded_Lf;


   end Compute_Scheduling_Period;





   procedure Compute_Response_Time (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 
      Rt : Response_Table;  

      Msg : Unbounded_String := Empty_String;  

   begin

      Compute_Response_Time(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Msg, Rt);
      Message:=Message & Lb_Task_Response_Time(Current_Language) & Msg &
         Unbounded_Lf;



      for I in 0..Response_Range(Get_Number_Of_Task_From_Processor(Sys.
               Tasks,A_Processor.Name)-1) loop
         if (Rt.Entries(I).Item /= null)
               then
            Message:=Message &  To_Unbounded_String("    ") & Rt.Entries(
               I).
               Item.Name & To_Unbounded_String(" => ") & Format(Rt.
               Entries(I).
               Data);

            if (Natural(Rt.Entries(I).Data) > Rt.Entries(I).Item.
                  Deadline)
                  then
               Message := Message & Lb_Comma & Lb_Check_Deadline(
                  Current_Language) & Rt.Entries(I).Item.Deadline'Img &
                  To_Unbounded_String(")");
            end if;


            Message:=Message &  Unbounded_Lf;

         end if;
      end loop;


   exception
      when Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
         Message:=Message & Lb_Compute_Error_4(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_5(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_6(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Processor_Utilization_Exceeded =>
         Message:=Message & Lb_Compute_Error_7(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Start_Time_Error =>
         Message:=Message & Lb_Compute_Error_12(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Offset_Error =>
         Message:=Message & Lb_Compute_Error_13(Current_Language) &
            Unbounded_Lf;

   end Compute_Response_Time;



   procedure Processor_Utilization (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 

      Result : Double;  
      Msg    : Unbounded_String;  
   begin

      Result:=Processor_Utilization_Over_Deadline(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);
      Message:= Message & Lb_Utilization_With_Deadline(Current_Language) &
         Format(Result) & Lb_Point & Unbounded_Lf;

      Result:=Processor_Utilization_Over_Period(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);
      Message:= Message & Lb_Utilization_With_Period(Current_Language) &
         Format(Result) &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & Unbounded_Lf;

      Bound_On_Processor_Utilization(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Result, Msg);
      Message:= Message & Lb_Utilization_Bound(Current_Language) & Format(
         Result) &
         Msg;


   exception
      when Task_Set.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_9(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_10(Current_Language) &
            Unbounded_Lf;


   end Processor_Utilization;



   procedure Set_Priorities (
         A_Window       : in     Root_Window_Access; 
         A_Scheduler    : in     Scheduler_Type;     
         Processor_Name : in     Unbounded_String    ) is 


      A_Rm_Scheduler : Rm_Scheduler;  
      A_Dm_Scheduler : Dm_Scheduler;  
      Ok             : Boolean      := True;  

   begin


      Write_Title(Root_Window, Lb_Processor(Current_Language)  &
         Processor_Name  & To_Unbounded_String(" : ") &
         Unbounded_Lf & Unbounded_Lf);


      begin

         if(A_Scheduler=Rate_Monotonic) then
            Set_Priority_According_To_Rm(A_Rm_Scheduler, Sys.Tasks,
               Processor_Name);
         else
            Set_Priority_According_To_Dm(A_Dm_Scheduler, Sys.Tasks,
               Processor_Name);
         end if;


         -- Do not forget to save project
         -- since tasks are modified !!
         --
         Is_New:=False;
         Is_Saved:=False;



      exception
         when Task_Set.Task_Must_Be_Periodic=>
            Ok:=False;

      end;

      Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf & Unbounded_Lf);
      if Ok
            then
         Write_Text(A_Window, Lb_Priorities_Are_Updated(Current_Language));

      else
         Write_Text(A_Window, Lb_Priorities_Error(Current_Language));
      end if;
      Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf & Unbounded_Lf);


   end Set_Priorities;



   procedure Chetto_Parameter_Modifications (
         A_Window           : in     Root_Window_Access; 
         Compute_Deadlines  : in     Boolean;            
         Compute_Priorities : in     Boolean;            
         Update_Tasks_Set   : in     Boolean             ) is 



      Tmp       : Tasks_Set;  
      A_Task    : Task_Ptr;  
      Iterator1 : Tasks_Iterator;  

   begin


      if Compute_Deadlines then

         Write_Title(Root_Window, Unbounded_Lf & Lb_Chetto_Deadline(
               Current_Language)  &
            To_Unbounded_String(" : ") &
            To_Unbounded_String("(") & Lb_See(Current_Language) &
            To_Unbounded_String("[5], [6])") &
            Unbounded_Lf & Unbounded_Lf);

      else
         Write_Title(Root_Window, Unbounded_Lf & Lb_Chetto_Priority(
               Current_Language)  &
            To_Unbounded_String(" : ") &
            To_Unbounded_String("(") & Lb_See(Current_Language) &
            To_Unbounded_String("[5], [6])") &
            Unbounded_Lf & Unbounded_Lf);
      end if;



      if Is_Acyclic(Sys.Dependencies)

            then


         -- If we must update parameters, we do not work on temporary tasks_set
         -- Call logic and display new parameters
         --
         if Update_Tasks_Set then
            Set_Parameters_According_To_Chetto(Sys.Dependencies, Sys.
               Tasks,
               Compute_Deadlines, Compute_Priorities);


            -- Do not forget to save project
            -- since tasks are modified !!
            --
            Is_New:=False;
            Is_Saved:=False;


            -- Display priorities/deadlines
            --
            Reset_Iterator(Iterator1);
            loop

               Current_Element(Sys.Tasks, A_Task, Iterator1);
               Write_Text(Root_Window,To_Unbounded_String("- ") & A_Task.
                  Name & To_Unbounded_String(" => "));
               if Compute_Deadlines
                     then
                  Write_Text(Root_Window, To_Unbounded_String(A_Task.
                        Deadline'Img) &
                     Unbounded_Lf);
               else
                  Write_Text(Root_Window, To_Unbounded_String(A_Task.
                        Priority'Img) &
                     Unbounded_Lf);
               end if;

               exit when Is_Last_Element(Sys.Tasks, Iterator1) = True;

               Next_Element(Sys.Tasks, Iterator1);
            end loop;


            if Compute_Priorities then
               Write_Text(A_Window, Lb_Priorities_Are_Updated(
                     Current_Language));
            else
               Write_Text(A_Window, Lb_Deadlines_Are_Updated(
                     Current_Language));
            end if;
            Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf &
               Unbounded_Lf);


         else
            Reset(Tmp);
            Duplicate(Sys.Tasks, Tmp);
            Set_Parameters_According_To_Chetto(Sys.Dependencies, Tmp,
               Compute_Deadlines, Compute_Priorities);


            -- Display priorities/deadlines
            --
            Reset_Iterator(Iterator1);
            loop

               Current_Element(Tmp, A_Task, Iterator1);
               Write_Text(Root_Window,To_Unbounded_String("- ") & A_Task.
                  Name & To_Unbounded_String(" => "));
               if Compute_Deadlines
                     then
                  Write_Text(Root_Window, To_Unbounded_String(A_Task.
                        Deadline'Img) &
                     Unbounded_Lf);
               else
                  Write_Text(Root_Window, To_Unbounded_String(A_Task.
                        Priority'Img) &
                     Unbounded_Lf);
               end if;

               exit when Is_Last_Element(Tmp, Iterator1) = True;

               Next_Element(Tmp, Iterator1);
            end loop;




         end if;



      else

         Write_Text(Root_Window, Lb_Cyclic_Error(Current_Language));


      end if;


      Write_Text(A_Window, Unbounded_Lf );



   end Chetto_Parameter_Modifications;



   procedure Compute_End_To_End_Response_Time (
         A_Window         : in     Root_Window_Access; 
         Update_Tasks_Set : in     Boolean;            
         One_Step         :        Boolean             ) is 


      Rt          : Response_Table;  
      Referencies : Unbounded_String := Empty_String;  
      Message     : Unbounded_String := Empty_String;  

   begin

      Initialize(Rt);

      Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf & Unbounded_Lf);


      if One_Step then
         Write_Title(A_Window, Lb_Minus &
            Lb_Jitter_From_Response_Time_One_Step(Current_Language) &
            Lb_Colon);
      else
         Write_Title(A_Window, Lb_Minus & Lb_Jitter_From_Response_Time(
               Current_Language) & Lb_Colon);
      end if;



      begin


         Compute_End_To_End_Response_Time(Sys.Dependencies, Sys.
            Processors, Sys.Tasks, One_Step, Update_Tasks_Set, Referencies,
            Rt);


         for I in 0..Rt.Nb_Entries-1 loop
            if (Rt.Entries(I).Item /= null)
                  then
               Message:=Message &  To_Unbounded_String("    ") & Rt.
                  Entries(
                  I).
                  Item.Name & To_Unbounded_String(" => ") & Format(Rt.
                  Entries(I).Data) & Lb_Comma & Lb_Jitter(
                  Current_Language) &
                  To_Unbounded_String("=")  & Periodic_Task_Ptr(Rt.
                  Entries(I).Item).Jitter'Img &
                  To_Unbounded_String(" ");

               if (Natural(Rt.Entries(I).Data) > Rt.Entries(I).Item.
                     Deadline)
                     then
                  Message := Message & Lb_Comma & Lb_Check_Deadline(
                     Current_Language) & Rt.Entries(I).Item.Deadline'Img;
                  Message:=Message  &
                     To_Unbounded_String(")");
               end if;

               Message:=Message & Unbounded_Lf;
            end if;
         end loop;



         if Update_Tasks_Set then

            Message:=Message & Lb_Jitters_Are_Updated(Current_Language) &
               Unbounded_Lf;

            -- Do not forget to save project
            -- since tasks are modified !!
            --
            Is_New:=False;
            Is_Saved:=False;


         end if;



      exception
         when Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
            Message:=Message & Lb_Compute_Error_4(Current_Language) &
               Unbounded_Lf;
         when Task_Set.Task_Must_Be_Periodic=>
            Message:=Message & Lb_Compute_Error_5(Current_Language) &
               Unbounded_Lf;
         when Task_Set.Task_Model_Error =>
            Message:=Message & Lb_Compute_Error_6(Current_Language) &
               Unbounded_Lf;
         when Scheduler.Processor_Utilization_Exceeded =>
            Message:=Message & Lb_Compute_Error_7(Current_Language) &
               Unbounded_Lf;
         when Task_Set.Start_Time_Error =>
            Message:=Message & Lb_Compute_Error_12(Current_Language) &
               Unbounded_Lf;
         when Task_Set.Offset_Error =>
            Message:=Message & Lb_Compute_Error_13(Current_Language) &
               Unbounded_Lf;


      end;



      Write_Text(A_Window, Referencies & Unbounded_Lf & Unbounded_Lf &
         Message);
      Write_Text(A_Window, Unbounded_Lf & Unbounded_Lf );





   end Compute_End_To_End_Response_Time;




begin

   Sched:=new Scheduling_Table;
   Initialize(Sched.all);

end Call_Scheduling_Logic;




