------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Tasks;
use Tasks;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Processor_Set;
use Processor_Set;
use Processor_Set.Generic_Processor_Set;

with Io_Definitions;
use Io_Definitions;

with Text_Io;
use Text_Io;

with Scheduler;
use Scheduler;
with Scheduler.Edf;
use Scheduler.Edf;

with Offsets;
use Offsets;

with Natural_Util;
use Natural_Util;


package body Generation is

   function Get_Rand_Period (
         Min  : in     Natural;  
         Max  : in     Natural;  
         Seed : in     Generator ) 
     return Natural is 

      Rand_Number : Uniformly_Distributed;  
      Period_Aux  : Double;  
      Period_Prod : Natural;  

   begin
      Rand_Number := Random(Seed); -- between 0 and 1
      Period_Aux := Double(Rand_Number) * (Double(Max) - Double(Min));
      Period_Aux := Period_Aux + Double(Min);
      Period_Prod := Natural(Period_Aux);
      return Period_Prod;
   end Get_Rand_Period;

   -- system 1 consumer/ N producers generation
   procedure Generation_1_N (
         Nb_Ordonnancement : Natural; 
         Period_Max        : Natural; 
         Per_Cons_Par      : Natural; 
         Save_Project      : Boolean  ) is 

      -- types
      type Set_Range is new Integer range 1 .. Nb_Ordonnancement; 
      type Tasks_Set_Set is array (Set_Range) of Tasks_Set; 
      type Roles_Set is array (Set_Range) of Buffer_Roles_Table; 
      type Size_Set is array (Set_Range) of Buffer_Size_Table; 
      type Bound_Set is array (Set_Range) of Natural; 

      -- variables
      Systems   : Tasks_Set_Set;  
      Offset    : Offset_Table;  
      Roles     : Roles_Set;  
      Sizes     : Size_Set;  
      Tptr      : Task_Ptr;  
      Cpu       : Unbounded_String := To_Unbounded_String ("cpu");  
      Nb_Prod   : Integer;  
      Per_Cons,  
      Per_Prod  : Natural;  
      Bounds    : Bound_Set;  
      Seed      : Generator;  
      Aux_Prod,  
      Flow_Aux,  
      Flow_Prod : Double;  
      Min,  
      Max       : Natural;  
      Cpus      : Processors_Set;  
      Edf       : Scheduler_Ptr;  

   begin


      Edf:= new Edf_Scheduler;
      Add_Processor(Cpus, Cpu, Edf);


      -- intialization
      Initialize(Offset);
      for I in Set_Range loop
         Initialize(Roles(I));
         Initialize(Sizes(I));
         Bounds(I) := 0;
      end loop;

      Nb_Prod := 1;
      Per_Cons := Per_Cons_Par; -- Periode of the consumer

      -- tasks sets generation
      for I in Set_Range loop

         Add_Task(Systems(I),Tptr, To_Unbounded_String("Cons"),
            Cpu, 0, 1, Per_Cons, Per_Cons, 0, 0, 1, Sched_Fifo, Offset);
         Roles(I).Entries(0).Item:=Tptr.Name;
         Roles(I).Entries(0).Data.Size:=1;
         Roles(I).Entries(0).Data.The_Role:=Consumer;

         -- variables initialization
         Reset(Seed);
         Max := Period_Max;
         Aux_Prod := 1.0/Double(Per_Cons);
         Per_Prod := Per_Cons;
         Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
            Max);
         Flow_Prod := 0.0;

         New_Line;
         New_Line;

         -- test of the maximum
         if (Flow_Aux < 0.0)
               then

            -- new maximum calculation
            Max := Nb_Prod * Per_Cons;
            Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
               Max);
            Put_Line ("No solution with period maximum :" &
               Period_Max'Img);
            Put_Line ("new  maximum :" & Max'Img);

         end if;

         -- Producers generation
         for J in 1 .. Buffer_Roles_Range(Nb_Prod) loop
            -- period of producers is based on the respect of the conservative data flow : the data produce is egal to the data consum
            if (J = Buffer_Roles_Range(Nb_Prod)) -- last producer
                  then
               Per_Prod := Natural(Double'Ceiling(1.0/Aux_Prod));
               Put_Line ("Period = " & Per_Prod'Img);
            else
               Min := Natural(Double'Ceiling(1.0/Flow_Aux));
               Per_Prod := Get_Rand_Period(Min,Max,Seed);
               Put_Line ("Period = " & Per_Prod'Img);

            end if;

            -- last producer period calculation
            Aux_Prod := Aux_Prod - 1.0/Double(Per_Prod);

            -- bandswith for next producers
            Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod) + 1.0/Double(
               Max);

            -- intermediary calculation of total data flow of the producers
            Flow_Prod := Flow_Prod + 1.0 / Double(Per_Prod);

            -- add the tasks to the system
            Add_Task(Systems(I),Tptr, (To_Unbounded_String("Prod") &
                  To_Unbounded_String(J'Img)),
               Cpu , 0, 1, Per_Prod, Per_Prod , 0, 0, 1,  Sched_Fifo,
               Offset);

            -- array of maximum generation
            Roles(I).Entries(J).Item:=Tptr.Name;
            Roles(I).Entries(J).Data.Size:=1;
            Roles(I).Entries(J).Data.The_Role:=Producer;

         end loop;

         -- display the total data of producers : checking
         Put_Line ("Producers data flow = " & Flow_Prod'Img);

         -- next system with one more producer
         Nb_Prod := Nb_Prod + 1;

      end loop;

      -- save the systems in files
      if (Save_Project = True) then
         for I in Set_Range loop
            Write_To_File (Systems(I),To_Unbounded_String("ProjGen") &
               Format(Natural(I)) & ".tsk", Save_In_New_File);
            Write_To_File (Cpus, To_Unbounded_String("ProjGen") &
               Format(Natural(I)) & ".proc", Save_In_New_File);
         end loop;
      end if;

   end Generation_1_N;

end Generation;
